# This module defines utilities for the cloud function api

from .utils import __API_VERSION__
import tempfile
import shutil
import subprocess
import glob
import os

__FUNCTION_MAPPING__ = {
    'create_submission': 'submit.py',
    'abort_submission': 'abort.py',
    'existence': 'internal.py',
    'redacted': 'internal.py',
    'check_abort': 'internal.py',
    'quotas': 'quotas.py',
    'register': 'register.py',
    'query_account': 'query.py'
}

def _deploy(function, endpoint, service_account=None, project=None, overload_version=None):
    if overload_version is None:
        overload_version = __API_VERSION__[endpoint]
    with tempfile.TemporaryDirectory() as tempdir:
        with open(os.path.join(tempdir, 'requirements.txt'), 'w') as w:
            w.write('google-auth\n')
            w.write('google-cloud-storage\n')
            w.write('google-cloud-kms\n')
            w.write('cryptography\n')
        shutil.copyfile(
            os.path.join(
                os.path.dirname(__file__),
                __FUNCTION_MAPPING__[function]
            ),
            os.path.join(tempdir, 'main.py')
        )
        shutil.copyfile(
            os.path.join(
                os.path.dirname(__file__),
                'utils.py'
            ),
            os.path.join(tempdir, 'utils.py')
        )
        cmd = 'gcloud {project} beta functions deploy {endpoint}-{version} --entry-point {function} --runtime python37 --trigger-http --source {path} {service_account}'.format(
            endpoint=endpoint,
            version=overload_version,
            function=function,
            path=tempdir,
            service_account='' if service_account is None else ('--service-account '+service_account),
            project='' if project is None else ('--project '+project)
        )
        print(cmd)
        subprocess.check_call(
            cmd,
            shell=True
        )



#!/usr/bin/python

# Copyright 2017 Google Inc.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file or at
# https://developers.google.com/open-source/licenses/bsd

# wdl_runner.py
#
# This script is a wrapper around Cromwell, which will:
#
# 1- take as input:
#     * a WDL file to describe a workflow
#     * a JSON file to describe the inputs
#     * a Google Cloud project ID (ignored if on GCE) to run in
#     * a GCS path to a Cromwell "working directory"
#     * a GCS path to an "output directory"
#
# 2- launch a local instance of Cromwell
#
# 3- submit the inputs to Cromwell to run the workflow
#
# 4- copy the Cromwell run metadata ('wdl_run_metadata.json') to the
#        output directory
#
# 5- copy the Cromwell outputs to the output directory
#
# Cromwell can be found at:
#     https://github.com/broadinstitute/cromwell

import argparse
import json
import logging
import os
import urllib2
import traceback

import cromwell_driver
import file_util
import sys_util
import wdl_outputs_util
import tempfile

WDL_RUN_METADATA_FILE = 'wdl_run_metadata.json'


def gce_get_metadata(path):
    """Queries the GCE metadata server the specified value."""
    req = urllib2.Request(
            'http://metadata/computeMetadata/v1/%s' % path,
            None, {'Metadata-Flavor': 'Google'})

    return urllib2.urlopen(req).read()


class Runner(object):

    def __init__(self, args, environ):
        self.args = args

        # Fetch all required environment variables, exiting if unset.
        self.environ = sys_util.copy_from_env(
                ['CROMWELL', 'CROMWELL_CONF'], environ)
        cromwell_conf = self.environ['CROMWELL_CONF']
        cromwell_jar = self.environ['CROMWELL']

        # Verify that the output directory is empty (or not there).
        if not file_util.verify_gcs_dir_empty_or_missing(self.args.output_dir):
            sys_util.exit_with_error(
                    "Output directory not empty: %s" % self.args.output_dir)

        # Plug in the working directory and the project id to the Cromwell conf
        self.fill_cromwell_conf(cromwell_conf,
                                                        self.args.working_dir, self.args.project)

        # Set up the Cromwell driver
        self.driver = cromwell_driver.CromwellDriver(cromwell_conf, cromwell_jar)
        if not self.args.batch:
            self.driver.start(3)

    def fill_cromwell_conf(self, cromwell_conf, working_dir, project):
        try:
            project_id = gce_get_metadata('project/project-id')

            if project and project != project_id:
                logging.warning("Overridding project ID %s with %s",
                                                project, project_id)

        except urllib2.URLError as e:
            logging.warning(
                    "URLError trying to fetch project ID from Compute Engine metdata")
            logging.warning(e)
            logging.warning("Assuming not running on Compute Engine")

            project_id = project

        new_conf_data = file_util.file_safe_substitute(cromwell_conf, {
                'project_id': project_id,
                'working_dir': working_dir
                })

        with open(cromwell_conf, 'wb') as f:
            f.write(new_conf_data)

    def copy_workflow_output(self, result):
        output_files = wdl_outputs_util.get_workflow_output(
                result['outputs'], self.args.working_dir)

        # Copy final output files (if any)
        logging.info("Workflow output files = %s", output_files)

        if output_files:
            file_util.gsutil_cp(output_files, "%s/" % self.args.output_dir)

    def copy_workflow_metadata(self, metadata, metadata_filename):

        logging.info("Copying run metadata to %s", self.args.output_dir)

        # Copy the run metadata
        with open(metadata_filename, 'w') as f:
            json.dump(metadata, f)

        file_util.gsutil_cp([metadata_filename], "%s/" % self.args.output_dir)


    def run(self):
        if self.args.batch is not None:
            try:
                file_util.gsutil_cp([os.environ['SUBMISSION_DATA_PATH']], 'submission.json')
                with open('submission.json') as r:
                    submission_data = json.load(r)
                runtime = submission_data['runtime'] if 'runtime' in submission_data else None
                self.driver.start(runtime['memory'] if runtime is not None else 3)
                logging.info("Starting batch request")
                # logging.info("SUBMITTING JOB " + repr((
                #     self.args.batch,
                #     self.args.wdl,
                #     self.args.workflow_inputs,
                #     self.args.workflow_options,
                #     runtime['batch_limit'] if runtime is not None else 250,
                #     runtime['query_limit'] if runtime is not None else 100
                # )))
                job_data = self.driver.batch(
                    self.args.batch,
                    self.args.wdl,
                    self.args.workflow_inputs,
                    self.args.workflow_options,
                    runtime['batch_limit'] if runtime is not None else 250,
                    runtime['query_limit'] if runtime is not None else 100
                )
                logging.info("Copying execution output file")
                with open('workflows.json', 'w') as w:
                    json.dump(
                        job_data,
                        w,
                        indent=2
                    )
                file_util.gsutil_cp(['workflows.json'], self.args.output_dir+'/')
                for data in job_data:
                    logging.info("Workflow %s exited with status %s" % (data['workflow_id'], data['workflow_status']))
                    # if data['workflow_output'] is not None:
                    #     metadata_filename = '%s.%s' % (
                    #         data['workflow_id'],
                    #         WDL_RUN_METADATA_FILE
                    #     )
                    #     with open(metadata_filename, 'w') as w:
                    #         json.dump(
                    #             data['workflow_metadata'],
                    #             w,
                    #             indent=2
                    #         )
                    #     file_util.gsutil_cp([metadata_filename], self.args.output_dir+'/')
                logging.info("Run complete")
                statuses = {wf['workflow_status'] for wf in job_data}
                # infer status then get submission object
                file_util.gsutil_cp([os.environ['SUBMISSION_DATA_PATH']], 'submission.json')
                with open('submission.json') as r:
                    submission_data = json.load(r)
                if len(statuses) == 1 and 'Succeeded' in statuses:
                    submission_data['status'] = 'Succeeded'
                elif 'Failed' in statuses:
                    submission_data['status'] = 'Failed'
                elif 'Aborted' in statuses:
                    submission_data['status'] = 'Aborted'
                else:
                    submission_data['status'] = 'Unknown'
                with open('submission.json', 'w') as w:
                    json.dump(submission_data, w, indent=2)
                file_util.gsutil_cp(['submission.json'], os.environ['SUBMISSION_DATA_PATH'])
                return
            except:
                logging.error("Batch submission failed: " + traceback.format_exc())
                file_util.gsutil_cp([os.environ['SUBMISSION_DATA_PATH']], 'submission.json')
                with open('submission.json') as r:
                    submission_data = json.load(r)
                submission_data['status'] = 'Error'
                with open('submission.json', 'w') as w:
                    json.dump(submission_data, w, indent=2)
                file_util.gsutil_cp(['submission.json'], os.environ['SUBMISSION_DATA_PATH'])
                return
        logging.info("starting")

        # Submit the job to the local Cromwell server
        (result, metadata) = self.driver.submit(
            self.args.wdl,
            self.args.workflow_inputs,
            self.args.workflow_options
        )
        logging.info(result)

        # Copy run metadata and output files to the output directory
        self.copy_workflow_metadata(metadata, WDL_RUN_METADATA_FILE)
        self.copy_workflow_output(result)

        logging.info("run complete")


def main():
    parser = argparse.ArgumentParser(description='Run WDLs')
    parser.add_argument('--wdl', required=True,
                                            help='The WDL file to run')
    parser.add_argument('--workflow-inputs', required=True,
                                            help='The workflow inputs (JSON) file')
    parser.add_argument('--workflow-options', required=False,
                                            help='The workflow options (JSON) file')
    parser.add_argument('--project', required=False,
                                            help='The Cloud project id')
    parser.add_argument('--working-dir', required=True,
                                            help='Location for Cromwell to put intermediate results.')
    parser.add_argument('--output-dir', required=True,
                                            help='Location to store the final results.')
    parser.add_argument('--batch', help="Submit batch request", default=None)


    args = parser.parse_args()

    # Sanitize the working and output paths
    args.working_dir.rstrip('/')
    args.output_dir.rstrip('/')

    # Write logs at info level
    FORMAT = '%(asctime)-15s %(module)s %(levelname)s: %(message)s'
    logging.basicConfig(level=logging.INFO, format=FORMAT)

    # Don't info-log every new connection to localhost, to keep stderr small.
    logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)
    logging.getLogger("requests.packages.urllib3.connectionpool").setLevel(
            logging.WARNING)

    runner = Runner(args, os.environ)
    runner.run()


if __name__ == '__main__':
    try:
        main()
    finally:
        print("<<<EOF>>>")



from google.cloud import storage
from google.cloud import kms_v1 as kms
import google.auth
from google.auth.transport.requests import AuthorizedSession
import google.oauth2.service_account
import google.oauth2.credentials
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec, padding, utils
import requests
from hashlib import md5, sha256
import os
import json
from urllib.parse import quote
import time

# TODO: Update all endpoints to v1 for release
__API_VERSION__ = {
    'submit': 'v1',
    'abort': 'v1',
    'register': 'v1',
    'signature': 'v1',
    'query': 'v1',
    'quotas': 'v1',
    'existence': 'frozen'
}
# The api version will allow versioning of cloud functions
# The patching system will be as follows:
# End users who update their lapdog will encounter errors because the cloud endpoints do not exist
# End users contact administrators, asking them to update
# Administrators can run lapdog patch {namespace} (to be implemented)
# Lapdog patch will run code from lapdog.patch.__project_admin_apply_patch
# This function will deploy new cloud functions and run any other arbitrary code
# Such as updating iam policy bindings or role permissions

__CROMWELL_TAG__ = 'v0.12.0'



def getblob(gs_path, credentials=None, user_project=None):
    """
    Return a GCP "blob" object for a given gs:// path.
    Path must start with "gs://".
    By default, uses the current application default credentials for authentication.
    Alternatively, you may provide a `google.auth.Credentials` object.
    When interacting with a requester pays bucket, you must set `user_project` to
    be the name of the project to bill for the data transfer fees
    """
    bucket_id = gs_path[5:].split('/')[0]
    bucket_path = '/'.join(gs_path[5:].split('/')[1:])
    return storage.Blob(
        bucket_path,
        storage.Client(credentials=credentials).bucket(bucket_id, user_project)
    )

def cors(*methods):
    """
    Wraps functions intended to handle inbound flask requests. The wrapped
    function will have CORS policies applied as follows:
    1) OPTIONS requests are returned with the accepted list of methods and localhost only origins
    2) Issues 405 responses to requests with bad methods
    3) Issues 403 responses to requests with bad origins
    4) Passes on requests to the decorated function only if they have an accepted method and localhost origin
    """
    def wrapper(func):
        def call(request):
            if request.method == 'OPTIONS':
                headers = {
                    'Access-Control-Allow-Origin': (
                        'http://localhost:4201' if not (
                            'Origin' in request.headers and request.headers['Origin'].startswith('http://localhost')
                        ) else request.headers['Origin']
                    ),
                    'Access-Control-Allow-Methods': ', '.join(methods),
                }
                return ('', 204, headers)
            elif request.method not in methods:
                return ('Not allowed', 405, {'Allow': ', '.join(methods)})
            elif 'Origin' in request.headers and not request.headers['Origin'].startswith('http://localhost'):
                return ('Forbidden', 403)
            result = list(func(request))
            print("RESULT", result)
            if isinstance(result[0], dict):
                result[0] = json.dumps(result[0])
            return tuple(result)
        return call
    return wrapper

def get_token_info(token):
    """
    Gets metadata for a given GCP access token.
    """
    try:
        data = requests.get('https://www.googleapis.com/oauth2/v1/tokeninfo?access_token='+token).json()
        return data
    except:
        return None

def ld_acct_in_project(account, ld_project=None):
    """
    Gets a user's Pet account within a given lapdog engine project.
    Project must be specified unless function is executing within a cloud function
    """
    if ld_project is None:
        ld_project = os.environ.get('GCP_PROJECT')
    return ('lapdog-'+ md5(account.encode()).hexdigest())[:30] + '@' + ld_project + '.iam.gserviceaccount.com'

def ld_meta_bucket_for_project(ld_project=None):
    """
    Gets the metadata bucket id for a given lapdog engine project.
    Project must be specified unless function is executing within a cloud function
    """
    if ld_project is None:
        ld_project = os.environ.get('GCP_PROJECT')
    return 'ld-metadata-'+md5(ld_project.encode()).hexdigest()

def ld_project_for_namespace(namespace):
    """
    Gets the lapdog engine project for a given namespace
    """
    prefix = ('ld-'+namespace)[:23]
    suffix = md5(prefix.encode()).hexdigest().lower()
    return prefix + '-' + suffix[:6]

def proxy_group_for_user(account):
    """
    Gets a firecloud proxy group name for a given account
    """
    info = account.split('@')
    return info[0]+'-'+info[1].split('.')[0]+'-lapdog'

def generate_user_session(token):
    """
    Generates a Google AuthorizedSession from a provided access token
    """
    return AuthorizedSession(
        google.oauth2.credentials.Credentials(token)
    )

def generate_default_session(scopes=None):
    return AuthorizedSession(google.auth.default(scopes=scopes)[0])

def generate_core_session():
    ld_project = os.environ.get('GCP_PROJECT')
    account = 'lapdog-worker@{}.iam.gserviceaccount.com'.format(ld_project)
    t = int(time.time())
    key_data = json.loads(getblob(
        'gs://{bucket}/auth_key.json'.format(
            bucket=ld_meta_bucket_for_project(ld_project)
        )
    ).download_as_string().decode())
    return AuthorizedSession(
        google.oauth2.service_account.Credentials.from_service_account_info(key_data).with_scopes([
            'https://www.googleapis.com/auth/cloud-platform',
            'https://www.googleapis.com/auth/genomics',
            'https://www.googleapis.com/auth/devstorage.read_write'
        ])
    )

def authenticate_bucket(bucket, namespace, workspace, session, core_session):
    """
    Used internally to authenticate a bucket for lapdog use
    """
    if os.environ.get('GCP_PROJECT') != ld_project_for_namespace(namespace):
        return False, 'This project is not responsible for the provided namespace'
    workspace_blob = getblob(
        'gs://{bucket}/DO_NOT_DELETE_LAPDOG_WORKSPACE_SIGNATURE'.format(
            bucket=bucket
        ),
        credentials=session.credentials
    )
    if not workspace_blob.exists():
        # Must generate signature
        print("Generating new workspace signature for", workspace)
        try:
            response = session.get(
                'https://api.firecloud.org/api/workspaces/{namespace}/{workspace}'.format(
                    namespace=namespace,
                    workspace=workspace
                ),
                timeout=5 # Long timeout because this is a 1-time operation
            )
            if response.status_code == 200:
                fc_bucket = response.json()['workspace']['bucketName']
                if fc_bucket != bucket:
                    return False, 'The provided bucket does not belong to this workspace'
                sign_object(
                    ('{}/{}/{}'.format(namespace, workspace, bucket)).encode(),
                    workspace_blob,
                    core_session.credentials
                )
                return True, fc_bucket
        except requests.ReadTimeout:
            pass
        return False, 'The Firecloud API is currently offline or took too long to respond'
    return (
        verify_signature(workspace_blob, ('{}/{}/{}'.format(namespace, workspace, bucket)).encode()),
        'Workspace authentication token had an invalid signature'
    )

def get_crypto_keys(name, credentials):
    return sorted(
        (key.name for key in kms.KeyManagementServiceClient(credentials=credentials).list_crypto_key_versions(name) if key.state == 1),
        key=lambda name:int(name.split('/')[-1]),
        reverse=True
    )

def sign_object(data, blob, credentials):
    blob.upload_from_string(
        kms.KeyManagementServiceClient(
            credentials=credentials
        ).asymmetric_sign(
            get_crypto_keys('projects/{ld_project}/locations/us/keyRings/lapdog/cryptoKeys/lapdog-sign'.format(
                ld_project=os.environ.get('GCP_PROJECT')
            ), credentials)[0],
            {'sha256': sha256(data).digest()}
        ).signature
    )

def verify_signature(blob, data, _is_blob=True):
    for key in get_crypto_keys('projects/{ld_project}/locations/us/keyRings/lapdog/cryptoKeys/lapdog-sign'.format(ld_project=os.environ.get('GCP_PROJECT')), generate_default_session().credentials):
        try:
            serialization.load_pem_public_key(
                kms.KeyManagementServiceClient().get_public_key(key).pem.encode('ascii'),
                default_backend()
            ).verify(
                blob.download_as_string() if _is_blob else blob,
                sha256(data).digest(),
                padding.PSS(
                    mgf=padding.MGF1(hashes.SHA256()),
                    salt_length=32
                ),
                utils.Prehashed(hashes.SHA256())
            )
            return True
        except InvalidSignature:
            pass
    return False

def validate_permissions(session, bucket):
    try:
        response = session.get(
            "https://www.googleapis.com/storage/v1/b/{bucket}"
            "/iam/testPermissions?permissions=storage.objects.list&"
            "permissions=storage.objects.get&permissions=storage.objects.create&"
            "permissions=storage.objects.delete".format(bucket=quote(bucket, safe='')),
        )
        if response.status_code == 200:
            data = response.json()
            if 'permissions' in data:
                return (
                    'storage.objects.list' in data['permissions'] or 'storage.objects.get' in data['permissions'],
                    'storage.objects.create' in data['permissions'] and 'storage.objects.delete' in data['permissions']
                )
            else:
                return False, False
        elif response.status_code == 401:
            data = response.json()
            if 'error' in data and 'message' in data['error']:
                if data['error']['message'] == 'Invalid Credentials':
                    return None, 'Expired Credentials'
                elif data['error']['message'] == 'Not Found':
                    return None, 'Bucket Not Found'
        return None, 'Invalid Credentials'
    except:
        return None, 'Unexpected Error'

def validate_submission_file(blob):
    if not blob.exists():
        return False, 'Submission.json file not found'
    blob.reload()
    if blob.size is None or blob.size >= 1073741824: # 1Gib
        return False, 'Submission.json file exceeded maximum allowed size'
    return True, 'OK'

def fetch_submission_blob(session, bucket, submission_id):
    return getblob(
        'gs://{bucket}/lapdog-executions/{submission_id}/submission.json'.format(
            bucket=bucket,
            submission_id=submission_id
        ),
        credentials=session.credentials
    )

def check_user_service_account(account, session=None):
    if session is None:
        session = generate_default_session([
            'https://www.googleapis.com/auth/cloud-platform',
            'https://www.googleapis.com/auth/iam'
        ])
    response = query_service_account(session, account)
    if response.status_code == 200:
        return response.json()
    return None

def query_service_account(session, account):
    return session.get(
        'https://iam.googleapis.com/v1/projects/{project}/serviceAccounts/{account}'.format(
            project=os.environ.get('GCP_PROJECT'),
            account=quote(account)
        )
    )

def update_iam_policy(session, grants, project=None):
    """
    Updates the IAM policy for a given project.
    `session` must be an AuthorizedSession for a user with sufficient permissions
    to update the project IAM policy.
    `grants` must be a dictionary of email:role, to grant in the project.
    Prepend each email with the account type ("user:john@example.net", "serviceAccount:accountName@project.iam.googleapis.com", "group:group@groups.google.com")
    `project` must be the lapdog engine project. Project is inferred automatically
    when running in a cloud function
    """
    if project is None:
        project = os.environ.get('GCP_PROJECT')
    policy = session.post(
        'https://cloudresourcemanager.googleapis.com/v1/projects/{project}:getIamPolicy'.format(
            project=project
        )
    ).json()
    for account_email, role in grants.items():
        target_role = 'projects/{}/roles/{}'.format(
            project,
            role
        )
        if target_role in {binding['role'] for binding in policy['bindings']}:
            for i, binding in enumerate(policy['bindings']):
                if binding['role'] == target_role:
                    policy['bindings'][i]['members'].append(
                        account_email
                    )
        else:
            policy['bindings'].append(
                {
                    'role': target_role,
                    'members': [
                        account_email
                    ]
                }
            )
    response = session.post(
        'https://cloudresourcemanager.googleapis.com/v1/projects/{project}:setIamPolicy'.format(
            project=project
        ),
        headers={'Content-Type': 'application/json'},
        json={
            "policy": policy,
            "updateMask": "bindings"
        }
    )
    return response.status_code == 200, response



"""
This module is intended to contain code required to patch a given namespace to the current version of lapdog
Lapdog patch will run __project_admin_apply_patch
This function may:
* Deploy new cloud functions to endpoints with updated versions
* Delete insecure old versions
* Modify project roles
* Update iam policy or bindings
* Regenerate signing keys
"""
from .utils import ld_project_for_namespace, __API_VERSION__, generate_user_session
from . import _deploy
from .. import __version__
from ..gateway import get_access_token
import sys
import crayons

# Notes for creating future patches
# 1) Import a Gateway and use get_endpoint to check for any redacted endpoints
#       Redact any endpoints which can be found
# 2) Always redeploy the current latest version of all endpoints/role definitions/api bindings
# 3) Mark the separate phases (iam, deploy, redact)

# Alternatively, write a function for each patch
# When running patches, check that the previous patch has been applied

def __project_admin_apply_patch(namespace):
    """
    PATCH SPEC: Beta -> V1
    """
    print("Patch version", __version__)
    print("Patching Namespace:", namespace)
    project = ld_project_for_namespace(namespace)
    print("Lapdog Project:", project)
    functions_account = 'lapdog-functions@{}.iam.gserviceaccount.com'.format(project)
    roles_url = "https://iam.googleapis.com/v1/projects/{project}/roles".format(
        project=project
    )
    print(crayons.black("Phase 1/3:", bold=True), "Update IAM Policies")
    user_session = generate_user_session(get_access_token())
    print("PATCH", roles_url+"/Core_account")
    response = user_session.patch(
        roles_url+"/Core_account",
        headers={
            'Content-Type': 'application/json'
        },
        json={
            "title": "Core_account",
            "includedPermissions": [
                "cloudkms.cryptoKeyVersions.useToSign",
                "cloudkms.cryptoKeyVersions.viewPublicKey",
                "cloudkms.cryptoKeyVersions.get",
                "cloudkms.cryptoKeyVersions.list",
                "resourcemanager.projects.get",
                "genomics.datasets.create",
                "genomics.datasets.delete",
                "genomics.datasets.get",
                "genomics.datasets.list",
                "genomics.datasets.update",
                "genomics.operations.cancel",
                "genomics.operations.create",
                "genomics.operations.get",
                "genomics.operations.list"
            ],
            "stage": "GA"
        }
    )
    if response.status_code != 200:
        print("(%d) : %s" % (response.status_code, response.text), file=sys.stderr)
        raise ValueError("Invalid response from Google API")
    print("PATCH", roles_url+"/Functions_account")
    response = user_session.patch(
        roles_url+"/Functions_account",
        headers={
            'Content-Type': 'application/json'
        },
        json={
            "title": "Functions_account",
            "includedPermissions": [
                "cloudkms.cryptoKeyVersions.viewPublicKey",
                "cloudkms.cryptoKeyVersions.get",
                "cloudkms.cryptoKeyVersions.list",
                "iam.serviceAccountKeys.create",
                "iam.serviceAccountKeys.delete",
                "iam.serviceAccountKeys.get",
                "iam.serviceAccountKeys.list",
                "iam.serviceAccounts.create",
                "iam.serviceAccounts.delete",
                "iam.serviceAccounts.get",
                "iam.serviceAccounts.getIamPolicy",
                "iam.serviceAccounts.list",
                "iam.serviceAccounts.setIamPolicy",
                "iam.serviceAccounts.update",
                "resourcemanager.projects.get",
                "genomics.operations.create",
                "genomics.operations.get",
                "genomics.operations.list",
                "resourcemanager.projects.getIamPolicy",
                "resourcemanager.projects.setIamPolicy",
                "compute.projects.get",
                "compute.regions.get"
            ],
            "stage": "GA"
        }
    )
    if response.status_code != 200:
        print("(%d) : %s" % (response.status_code, response.text), file=sys.stderr)
        raise ValueError("Invalid response from Google API")
    print(crayons.black("Phase 2/3:", bold=True), "Deploy Cloud API Updates")
    print("Patching submit from beta -> v1")
    _deploy('create_submission', 'submit', functions_account, project)
    print("Patching abort from beta -> v1")
    _deploy('abort_submission', 'abort', functions_account, project)
    print("Patching signature from beta -> v1")
    _deploy('check_abort', 'signature', functions_account, project)
    print("Patching register from beta -> v1")
    _deploy('register', 'register', functions_account, project)
    print("Patching query from beta -> v1")
    _deploy('query_account', 'query', functions_account, project)
    print("Patching quotas from beta -> v1")
    _deploy('quotas', 'quotas', functions_account, project)
    print(crayons.black("Phase 3/3:", bold=True), "Redact Insecure Cloud API Endpoints")
    print(crayons.green("No Redactions"))



from google.cloud import storage
import subprocess
import yaml
from io import StringIO, BytesIO
import json
import time
import datetime
import threading
import os
from functools import lru_cache
import contextlib
import re
import select
from .cache import cache_fetch, cache_write, cached, cache_path
from .cloud.utils import getblob
from .gateway import Gateway
import traceback
import sys
import threading
from hashlib import md5
import warnings
from agutil import ActiveTimeout, TimeoutExceeded, context_lock

# Label filter format: labels.(label name)=(label value)

timestamp_formats = (
    '%Y-%m-%dT%H:%M:%SZ',
    '%Y-%m-%dT%H:%MZ'
)
utc_offset = datetime.datetime.fromtimestamp(time.time()) - datetime.datetime.utcfromtimestamp(time.time())

def sleep_until(dt):
    sleep_time = (dt - datetime.datetime.now()).total_seconds()
    if sleep_time > 0:
        time.sleep(sleep_time)

def parse_time(timestamp):
    """
    Returns a datetime.datetime object from a given UTC timestamp.
    Parses timestamps in multiple known GCP timestamp formats
    """
    #2019-02-04T21:27Z
    # no seconds?
    err = None
    for timestamp_format in timestamp_formats:
        try:
            return datetime.datetime.strptime(
                (timestamp.split('.')[0]+'Z').replace('ZZ', 'Z'),
                timestamp_format
            )
        except ValueError as e:
            err = e
    raise err

# individual VMs can be tracked viua a combination of labels and workflow meta
# Workflows are reported by the cromwell_driver output and then VMs can be tracked
# A combination of the Workflow label and the call- label should be able to uniquely identify tasks
# Scatter jerbs will have colliding VMs, but can that will have to be handled elsewhere
# Possibly, the task start pattern will contain multiple operation starts (or maybe shard-ids are given as one of the digit fields)

workflow_dispatch_pattern = re.compile(r'Workflows(( [a-z0-9\-]+,?)+) submitted.')
workflow_start_pattern = re.compile(r'WorkflowManagerActor Successfully started WorkflowActor-([a-z0-9\-]+)')
task_start_pattern = re.compile(r'\[UUID\((\w{8})\)(\w+)\.(\w+):(\w+):(\d+)\]: job id: ((?:projects/.+/)?operations/\S+)')
#(short code), (workflow name), (task name), (? 'NA'), (call id), (operation)
msg_pattern = re.compile(r'\[UUID\((\w{8})\)\]')
#(short code), message
fail_pattern = re.compile(r"ERROR - WorkflowManagerActor Workflow ([a-z0-9\-]+) failed \(during *?\): (.+)")
#(long id), (opt: during status), (failure message)
status_pattern = re.compile(r'PipelinesApiAsyncBackendJobExecutionActor \[UUID\(([a-z0-9\-]+)\)(\w+)\.(\w+):(\w+):(\d+)]: Status change from (.+) to (.+)')
#(short code), (workflow name), (task name), (? 'NA'), (call id), (old status), (new status)

instance_name_pattern = re.compile(r'instance(?:Name)?:\s+(.+)')

disk_pattern = re.compile(r'local-disk (\d+) (HDD|SSD)')

mtypes = {
    'n1-standard-%d'%(2**i): (0.0475*(2**i), 0.01*(2**i)) for i in range(7)
}
mtypes.update({
    'n1-highmem-%d'%(2**i): (.0592*(2**i), .0125*(2**i)) for i in range(1,7)
})
mtypes.update({
    'n1-highcpu-%d'%(2**i): (.03545*(2**i), .0075*(2**1)) for i in range(1,7)
})
mtypes.update({
    'f1-micro': (0.0076, 0.0035),
    'g1-small': (0.0257, 0.007)
})

core_price = (0.031611, 0.006655)
mem_price = (0.004237, 0.000892)
extended_price = (0.009550, 0.002014)

def get_hourly_cost(mtype, preempt=False):
    """
    Returns the cost/hour of a given GCP machine type.
    All machine types supported, including predefined n1, f1, and g1 instance families
    As well as custom and custom extended memory instances.
    If `preempt` is True, hourly cost will reflect preemptible price model
    """
    try:
        if mtype in mtypes:
            return mtypes[mtype][int(preempt)]
        else:
            if mtype.endswith('-ext'):
                mtype = mtype[:-4]
            custom, cores, mem = mtype.split('-')
            return (core_price[int(preempt)]*int(cores)) + (mem_price[int(preempt)]*max(int(mem), int(cores) * 1024 * 6.5)/1024) + max(0, extended_price[int(preempt)]*(int(mem)-(int(cores)*1024*6.5))/1024)
    except:
        traceback.print_exc()
        print(mtype, "unknown machine type")
        return 0

def get_cromwell_type(runtime):
    if runtime['memory'] <= 3:
        return 'n1-standard-1'
    return 'custom-2-%d' % (1024 * runtime['memory'])

class NoSuchSubmission(Exception):
    pass

class Recall(object):
    value = None
    def apply(self, value):
        self.value = value
        return value

def safe_getblob(gs_path):
    """
    A wrapper to getblob which raises a FileNotFoundError if the gs:// path does not exist
    Use when reading a file
    """
    blob = getblob(gs_path)
    if not blob.exists():
        raise FileNotFoundError("No such blob: "+gs_path)
    return blob

def do_select(reader, t):
    if isinstance(reader, BytesIO):
        # print("Bytes seek")
        current = reader.tell()
        reader.seek(0,2)
        end = reader.tell()
        if current < end:
            # print("There are", end-current, "bytes")
            reader.seek(current, 0)
            return [[reader]]
        return [[]]
    else:
        return select.select([reader], [], [], t)


class Call(object):
    """
    Class represents a given Task in a workflow
    """
    status = '-'
    last_message = ''
    def __init__(self, path, task, attempt, operation):
        self.path = path
        self.task = task
        self.attempt = attempt
        self.operation = operation

    @property
    @cached(10)
    def return_code(self):
        """
        Property. The return code from the Task's script.
        10 second cache
        """
        try:
            blob = safe_getblob(os.path.join(self.path, 'rc'))
            return int(blob.download_as_string().decode())
        except FileNotFoundError:
            return None
        except ValueError:
            return None

    @property
    @cached(10)
    def runtime(self):
        """
        Property. The time in hours that the Task was running.
        10 second cache
        """
        try:
            data = get_operation_status(self.operation)
            delta = (
                parse_time(data['metadata']['endTime'])
                if 'endTime' in data['metadata']
                else datetime.datetime.utcnow()
            ) - parse_time(data['metadata']['startTime'])
            return delta.total_seconds() / 3600
        except:
            traceback.print_exc()
            return 0

@cached(10)
def get_operation_status(opid, parse=True, fmt='json'):
    """
    Fetches the metadata for a given GCP operation ID.
    If `parse` is True (default) the json string will be parsed into a python dict.
    `fmt` sets the download format. Do not change unless you have reason to download
    in a different data format.
    10 second cache.
    """
    text = cache_fetch('operation', opid)
    if text is None:
        text = subprocess.run(
            'gcloud alpha genomics operations describe --format=%s %s' % (
                fmt,
                opid
            ),
            shell=True,
            executable='/bin/bash',
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        ).stdout.decode()
        if not parse:
            return text
        try:
            data = yaml.load(StringIO(text))
            if 'done' in data and data['done']:
                cache_write(text, 'operation', opid)
        except yaml.scanner.ScannerError:
            if 'Permission denied' in text:
                raise ValueError("Permission Denied")
            print(text)
            raise
    else:
        data = yaml.load(StringIO(text))
    if not parse:
        return text
    return data

def abort_operation(opid):
    return subprocess.run(
        'yes | gcloud alpha genomics operations cancel %s' % (
            opid
        ),
        shell=True,
        executable='/bin/bash',
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

def kill_machines(instance_name):
    # subprocess.check_call(
    #     'gcloud compute instances stop %s --async' % instance_name,
    #     shell=True
    # )
    # machines = subprocess.run(
    #     'gcloud compute instances list --filter="labels.%s"' % (
    #         label
    #     ),
    #     shell=True,
    #     executable='/bin/bash',
    #     stdout=subprocess.PIPE,
    #     stderr=subprocess.STDOUT,
    # ).stdout.decode().split('\n')
    # if len(machines) > 1:
    #     machines = ' '.join(line.split()[0] for line in machines[1:] if len(line.strip()))
    # if len(machines):
    # Use Popen instead of run because deleting instances is a slow, blocking operation
    return subprocess.Popen(
        'yes | gcloud compute instances delete %s' % (
            instance_name
        ),
        shell=True,
        executable='/bin/bash',
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

class CommandReader(object):
    """
    Reads buffered output from a subprocess command.
    __init__ takes identical arguments to subprocess.Popen
    Call readline() to get one line of text
    """
    def __init__(self, cmd, *args, __insert_text=None, **kwargs):
        r,w = os.pipe()
        r2,w2 = os.pipe()
        if '__insert_text' in kwargs:
            __insert_text = kwargs['__insert_text']
            del kwargs['__insert_text']
        if __insert_text is not None:
            os.write(w, __insert_text)
        # self.proc = subprocess.Popen(cmd, *args, stdout=w, stderr=w, stdin=r2, universal_newlines=False, **kwargs)
        self.proc = subprocess.Popen(cmd, *args, stdout=w, stderr=subprocess.DEVNULL, stdin=r2, universal_newlines=False, **kwargs)
        self.reader = open(r, 'rb')
        self.buffer = b''

    def close(self, *args, **kwargs):
        """
        Closes the input stream and kills the underlying process
        """
        self.reader.close(*args, **kwargs)
        if self.proc.returncode is None:
            self.proc.kill()

    def readline(self, length=256, *args, **kwargs):
        """
        Reads a single line of text from the underlying process.
        `length` specifies the number of bytes that are read in each chunk while
        waiting for a complete line of text
        """
        while b'\n' not in self.buffer:
            self.buffer += os.read(self.reader.fileno(), length)
        self.buffer = self.buffer.split(b'\n')
        output = self.buffer.pop(0)
        self.buffer = b'\n'.join(self.buffer)
        return output

    def __getattr__(self, attr):
        if hasattr(self.reader, attr):
            return getattr(self.reader, attr)
        raise AttributeError("No such attribute '%s'" % attr)

    def __del__(self):
        self.close()

## TODO:
## 1) Controller:
##      Cache a cromwell reader and text from the most recent submission
##      On submission change, close the reader and reset text
##      On fetch: use select to check for input (50ms timeout) then read all
##      Optional line_offset input returns that line and all after
## 2) Adapter:

class SubmissionAdapter(object):
    """
    Represents a single Lapdog Submission.
    Use to interact with the submission, including fetching status and workflows
    or aborting a running submission
    """
    def __init__(self, bucket, submission, gateway=None):
        """
        Constructs the adapter. Requires the bucket id for the workspace and the
        submission id for the submission. To save time, you may also provide the
        Gateway object for this namespace, otherwise a new Gateway will be constructed.
        Downloads and parses the submission.json file for this submission.
        submission.json files are cached if the submission is done.
        """
        # print("Constructing adapter")
        self.path = os.path.join(
            'gs://'+bucket,
            'lapdog-executions',
            submission
        )
        gs_path = os.path.join(
            self.path,
            'submission.json'
        )
        self.data = cache_fetch('submission-json', bucket, submission)
        _do_cache_write = False
        if self.data is None:
            try:
                self.data = safe_getblob(gs_path).download_as_string().decode()
                self._cached = False
                _do_cache_write = True
            except FileNotFoundError as e:
                raise NoSuchSubmission from e
        else:
            self._cached = True
        self.data = json.loads(self.data)
        if 'operation' not in self.data:
            print("<GATEWAY DEV> Delete submission", submission)
        self.workspace = self.data['workspace']
        self.namespace = self.data['namespace']
        self.identifier = self.data['identifier']
        self.operation = self.data['operation'] if 'operation' in self.data else 'NULL'
        self.raw_workflows = self.data['workflows']
        self.gateway = Gateway(self.namespace) if gateway is None else gateway
        self.workflow_mapping = {}
        self.thread = None
        self.bucket = bucket
        self.submission = submission
        self.workflows = {}
        self._internal_reader = None
        self.bucket = bucket
        self._update_mask = set()
        self.update_lock = threading.Lock()
        if _do_cache_write and not self.live:
            cache_write(json.dumps(self.data), 'submission-json', bucket, submission)
            self._cached = True

    def _init_workflow(self, short, key=None, long_id=None):
        if short not in self.workflows:
            self.workflows[short] = WorkflowAdapter(short, self.path, key, long_id)
        return self.workflows[short]

    def cost(self):
        """
        Estimates the cost of a submission.
        If the submission is not running, parse the workflow.json file to get the
        exact start/stop times for every call of every workflow.
        If the submission is running, workflow.json was not found, or any error was encountered during the above computation:
        Update the adapter, and estimate cost by parsing the operation metadata for
        each Call in the submission. This pathway stops if the computation takes more than 60 seconds.
        Cost is cached if the submission is not running and the first cost estimation pathway finishes.
        Returns a dictionary with the elapsed wall time, total cpu time, total cost,
        and overhead cost of the cromwell server
        """
        try:
            if not self.live:
                stored_cost = cache_fetch('submission', self.namespace, self.workspace, self.submission, dtype='cost')
                if stored_cost is not None:
                    return json.loads(stored_cost)
                try:
                    workflow_metadata = json.loads(safe_getblob(
                        os.path.join(
                            self.path,
                            'results',
                            'workflows.json'
                        )
                    ).download_as_string())
                    cost = 0
                    total = 0
                    status = self.status
                    if 'metadata' in status and 'startTime' in status['metadata']:
                        maxTime = (
                            parse_time(status['metadata']['endTime']) if 'endTime' in status['metadata']
                            else datetime.datetime.utcnow()
                        ) - parse_time(status['metadata']['startTime'])
                        maxTime = maxTime.total_seconds() / 3600
                    for wf in workflow_metadata:
                        if wf['workflow_metadata'] is not None and 'calls' in wf['workflow_metadata']:
                            for calls in wf['workflow_metadata']['calls'].values():
                                for call in calls:
                                    if 'end' in call:

                                        delta = (
                                            # Stupid horrible ugly timestampt
                                            parse_time(call['end']) - parse_time(call['start'])
                                        )
                                        delta = delta.total_seconds() / 3600
                                        # if delta > maxTime:
                                        #     maxTime = delta
                                        total += delta
                                        if 'jes' in call and 'machineType' in call['jes']:
                                            cost += delta * get_hourly_cost(
                                                call['jes']['machineType'].split('/')[-1],
                                                'preemptible' in call and call['preemptible']
                                            )
                                        if 'runtimeAttributes' in call and 'disks' in call['runtimeAttributes'] and disk_pattern.match(call['runtimeAttributes']['disks']):
                                            result = disk_pattern.match(call['runtimeAttributes']['disks'])
                                            if result.group(2) == 'HDD':
                                                # runtime * 1hr/1month * monthly price/GB * GB
                                                cost += delta * 0.0013698624132109968 * .04 * float(result.group(1))
                                            elif result.group(2) == 'SSD':
                                                cost += delta * 0.0013698624132109968 * .17 * float(result.group(1))
                                            if 'bootDiskSizeGb' in call['runtimeAttributes']:
                                                cost += delta * 0.0013698624132109968 * .04 * float(call['runtimeAttributes']['bootDiskSizeGb'])
                    cost += maxTime * get_hourly_cost(
                        get_cromwell_type(self.data['runtime']) if 'runtime' in self.data else 'n1-standard-1',
                        False
                    )
                    cost = int(cost * 100) / 100
                    result = {
                        'clock_h': maxTime,
                        'cpu_h': total,
                        'est_cost': cost,
                        'cromwell_overhead': int(maxTime * get_hourly_cost(
                            get_cromwell_type(self.data['runtime']) if 'runtime' in self.data else 'n1-standard-1',
                            False
                        ) * 100) / 100
                    }
                    cache_write(json.dumps(result), 'submission', self.namespace, self.workspace, self.submission, dtype='cost')
                    return result
                except:
                    # Try the slow route
                    print(traceback.format_exc(), file=sys.stderr)
                    print("Attempting slow cost calculation")
                    pass
            try:
                cost = 0
                maxTime = 0
                total = 0
                with ActiveTimeout(60) as timer:
                    self.update()
                    timer.update()
                    for wf in self.workflows.values():
                        for call in wf.calls:
                            timer.update()
                            try:
                                call = get_operation_status(call.operation)
                                delta = (
                                    parse_time(call['metadata']['endTime'])
                                    if 'endTime' in call['metadata']
                                    else datetime.datetime.utcnow()
                                ) - parse_time(call['metadata']['startTime'])
                                delta = delta.total_seconds() / 3600
                                # if delta > maxTime:
                                #     maxTime = delta
                                total += delta
                                if 'jes' in call and 'machineType' in call['jes']:
                                    cost += delta * get_hourly_cost(
                                        call['jes']['machineType'].split('/')[-1],
                                        'preemptible' in call['metadata']['request']['pipelineArgs']['resources']
                                        and call['metadata']['request']['pipelineArgs']['resources']['preemptible']
                                    )
                            except KeyError:
                                pass
                            except yaml.scanner.ScannerError:
                                print("Operation", call.operation, "had invalid operation metadata")
            except TimeoutExceeded:
                print("Took longer than 60 seconds to load workflow costs. Skipping")
            status = self.status
            if 'metadata' in status and 'startTime' in status['metadata']:
                maxTime = (
                    parse_time(status['metadata']['endTime']) if 'endTime' in status['metadata']
                    else datetime.datetime.utcnow()
                ) - parse_time(status['metadata']['startTime'])
                maxTime = maxTime.total_seconds() / 3600
            cost += maxTime * get_hourly_cost(
                get_cromwell_type(self.data['runtime']) if 'runtime' in self.data else 'n1-standard-1',
                False
            )
            cost = int(cost * 100) / 100
            return {
                'clock_h': maxTime,
                'cpu_h': total,
                'est_cost': cost,
                'cromwell_overhead': int(maxTime * get_hourly_cost(
                    get_cromwell_type(self.data['runtime']) if 'runtime' in self.data else 'n1-standard-1',
                    False
                ) * 100) / 100
            }
        except:
            traceback.print_exc()
            print("STRING FORMAT ERROR", self.submission)
            return {
                'clock_h': 0,
                'cpu_h': 0,
                'est_cost': 0,
                'cromwell_overhead': 0
            }

    @cached(90)
    def update(self, timeout=-1):
        """
        Fetch latest data from Cromwell Server to update the adapter.
        Specify a positive number for `timeout` to set the maximum time to wait
        to acquire the lock on the underlying data stream.
        If acquiring the lock takes more than 10 seconds, do not update the adapter.
        This is because another thread must have just finished an update operation,
        so there is no need to immediately re-update.
        90 second cache.
        """
        ctime = time.monotonic()
        try:
            with context_lock(self.update_lock, timeout):
                if time.monotonic() - ctime > 10:
                    # We've been waiting a while already
                    # no reason to re-update, if another thread just spent a while
                    return
                # if self._internal_reader is None:
                #     self._internal_reader = self.read_cromwell(_do_wait=self.live)
                event_stream = []
                message = ''
                reader = self.read_cromwell(_do_wait=self.live)
                while len(do_select(reader, 1)[0]):
                    message = reader.readline().decode().strip()
                    code = md5(message.encode()).digest()
                    matcher = Recall()
                    if matcher.apply(workflow_dispatch_pattern.search(message)):
                        if code in self._update_mask:
                            continue
                        self._update_mask.add(code)
                        # This event helps establish the order of workflows
                        # The message is posted with the exact order of workflow-ids
                        # Which will match the order of entities dispatched in the submission
                        # It creates a new workflow object and inserts the mapping
                        # of key->id into the table
                        ids = [
                            wf_id.strip() for wf_id in
                            matcher.value.group(1).split(',')
                        ]
                        # print(len(ids), 'workflow(s) dispatched')
                        for long_id, data in zip(ids, self.raw_workflows[len(self.workflow_mapping):]):
                            self._init_workflow(
                                long_id[:8],
                                data['workflowOutputKey'],
                                long_id
                            )
                            self.workflow_mapping[data['workflowOutputKey']] = long_id
                    if matcher.apply(workflow_start_pattern.search(message)):
                        if code in self._update_mask:
                            continue
                        self._update_mask.add(code)
                        # This event captures a start message for a workflow which somehow
                        # was not captured by the above dispatch event
                        long_id = matcher.value.group(1)
                        if long_id not in {*self.workflow_mapping.values()}:
                            warnings.warn("Unexpected state")
                            traceback.print_stack()
                            idx = len(self.workflow_mapping)
                            data = self.raw_workflows[idx]
                            self._init_workflow(
                                long_id[:8],
                                data['workflowOutputKey'],
                                long_id
                            )
                            self.workflow_mapping[data['workflowOutputKey']] = long_id

                    elif matcher.apply(task_start_pattern.search(message)):
                        if code in self._update_mask:
                            continue
                        self._update_mask.add(code)
                        short = matcher.value.group(1)
                        wf = matcher.value.group(2)
                        task = matcher.value.group(3)
                        na = matcher.value.group(4)
                        call = int(matcher.value.group(5))
                        operation = matcher.value.group(6)
                        self._init_workflow(short).handle(
                            'task',
                            wf,
                            task,
                            na,
                            call,
                            operation
                        )
                    elif matcher.apply(msg_pattern.search(message)):
                        if code in self._update_mask:
                            continue
                        self._update_mask.add(code)
                        self._init_workflow(matcher.value.group(1)).handle(
                            'message',
                            matcher.value.string
                        )
                    elif matcher.apply(fail_pattern.search(message)):
                        if code in self._update_mask:
                            continue
                        self._update_mask.add(code)
                        long_id = matcher.value.group(1)
                        self._init_workflow(long_id[:8]).handle(
                            'fail',
                            matcher.groups()[-1]
                        )
                    elif matcher.apply(status_pattern.search(message)):
                        if code in self._update_mask:
                            continue
                        self._update_mask.add(code)
                        short = matcher.value.group(1)
                        task = matcher.value.group(3)
                        call = int(matcher.value.group(5))
                        old = matcher.value.group(6)
                        new = matcher.value.group(7)
                        self._init_workflow(short).handle(
                            'status',
                            task,
                            call,
                            old,
                            new
                        )
        except TimeoutExceeded:
            pass
            # else:
            #     print("NO MATCH:", message)

    def update_data(self):
        """
        Fetches latest submission data
        """
        if self._cached:
            return
        with self.update_lock:
            gs_path = os.path.join(
                self.path,
                'submission.json'
            )
            self.data = json.loads(safe_getblob(gs_path).download_as_string().decode())
            if 'operation' not in self.data:
                print("<GATEWAY DEV> Delete submission", submission)
            self.workspace = self.data['workspace']
            self.namespace = self.data['namespace']
            self.identifier = self.data['identifier']
            self.operation = self.data['operation'] if 'operation' in self.data else 'NULL'

    def abort(self):
        """
        Abort a running workflow.
        If the current status is "Running", send a soft abort, giving the Cromwell
        server time to cancel running workflows and clean up. Set status to "Aborting".

        If the abort procedure is taking too long, you may abort again:
        If the current status is "Aborting", send a hard abort, immediately killing
        the Cromwell server. This may leave running workflows in an orphan state
        where they continue to run until completion with no way to halt them.
        Forcefully set status to "Aborted".
        """
        # self.update()
        # FIXME: Once everything else works, see if cromwell labels work
        # At that point, we can add an abort here to kill everything with the id
        status = self.status
        if 'done' in status and status['done']:
            return
        try:
            # For now, use abort key
            result = self.gateway.abort_submission(
                self.bucket,
                self.data['submission_id'],
                self.data['status'] == 'Aborting' # Hard abort if we're already aborting
            )
            if result is not None:
                # Failed
                warnings.warn(
                    "Submission Abort did not complete. Some workflows may continue running unsupervised",
                    RuntimeWarning
                )
                print(result.text, file=sys.stderr)
        except:
            traceback.print_exc()
            warnings.warn(
                "Submission Abort did not complete. Some workflows may continue running unsupervised",
                RuntimeWarning
            )
        finally:
            gs_path = os.path.join(
                self.path,
                'submission.json'
            )
            getblob(gs_path).upload_from_string(
                json.dumps(
                    {
                        **self.data,
                        **{'status': 'Aborted' if self.data['status'] == 'Aborting' else 'Aborting'}
                    }
                )
            )




    @property
    @cached(5)
    def status(self):
        """
        Property. Get the operation status
        5 second cache
        """
        # print("READING ADAPTER STATUS")
        return get_operation_status(self.operation)

    @property
    @cached(10)
    def live(self):
        """
        Property. Reports if the submission is active or not
        10 second cache
        """
        status = self.status
        return not ('done' in status and status['done'])

    def read_cromwell(self, _do_wait=True):
        """
        Attempts to open a data stream to the cromwell server.
        Currently, the data stream is a BytesIO object of the most recent log
        output.
        In the future, submissions will use an Stackdriver event stream
        and fallback to file logs.
        If the submission has been recently started, this blocks for ~2 minutes
        """
        status = self.status # maybe this shouldn't be a property...it takes a while to load
        cromwell_text = cache_fetch('submission', self.namespace, self.workspace, self.submission, dtype='cromwell', decode=False)
        if cromwell_text is not None:
            return BytesIO(cromwell_text)
        while 'metadata' not in status or ('startTime' not in status['metadata'] and 'endTime' not in status['metadata']):
            status = self.status
            time.sleep(1)
        if _do_wait:
            sleep_until(
                parse_time(status['metadata']['startTime'])
                + utc_offset
                + datetime.timedelta(seconds=120)
            )
        stdout_blob = getblob(os.path.join(
            'gs://'+self.bucket,
            'lapdog-executions',
            self.submission,
            'logs',
            'stdout.log'
        ))
        if stdout_blob.exists():
            log_text = stdout_blob.download_as_string()
            cache_write(log_text, 'submission', self.namespace, self.workspace, self.submission, dtype='cromwell', decode=False)
            return BytesIO(log_text)
        stdout_blob = getblob(os.path.join(
            'gs://'+self.bucket,
            'lapdog-executions',
            self.submission,
            'logs',
            'pipeline-stdout.log'
        ))
        if stdout_blob.exists():
            log_text = stdout_blob.download_as_string()
            return BytesIO(log_text)
        stdout_blob = getblob(os.path.join(
            'gs://'+self.bucket,
            'lapdog-executions',
            self.submission,
            'logs',
            self.operation[11:]+'-stdout.log'
        ))
        if stdout_blob.exists():
            log_text = stdout_blob.download_as_string()
            cache_write(log_text, 'submission', self.namespace, self.workspace, self.submission, dtype='cromwell', decode=False)
            return BytesIO(log_text)
        if self.data['status'] != 'Error' and 'done' in status and status['done']:
            # If we get here, the submission is done, but there were no logs
            self.data['status'] = 'Error'
            cache_write(json.dumps(self.data), 'submission-json', self.bucket, self.submission)
            gs_path = os.path.join(
                self.path,
                'submission.json'
            )
            getblob(gs_path).upload_from_string(json.dumps(self.data).encode())
        return BytesIO(b'')


    # def get_workflows(self, workflows):
    #
    #     #LINK via ordering. we know the order in which workflows were submitted
    #     #And so we should know the order in which they are returned
    #     # This, at the very least, informs the adapter how many workflows to expect
    #     # The workflow_start_pattern will inform cromwell of when each workflow checks in
    #     # Depending on the data available in the status_json we may or may not be able to
    #     # Link workflows at this point
    #     # Alternatively, we can sniff the logs in from the workflow itself
    #     # Input keys are derived entirely from values, not variable names, so we
    #     #   might have enough data to link workflows that way
    #
    #     # This should start a monitoring thread to watch for patterns in the cromwell logs
    #     # Then, relevant information is logged to this object or child workflow Adapters
    #     pass


class WorkflowAdapter(object):
    """
    Represents a single workflow.
    It is best not to manipulate this object directly as it is very closely connected
    to its parent SubmissionAdapter.

    Do not call the `handle` method or any of the `on_...` methods as this may result
    in your WorkflowAdapter being stuck in an invalid state
    """
    # this adapter needs to be initialized from an input key and a cromwell workflow id (short or long)
    # At first, dispatching events fills the replay buffer
    # when the workflow is started, the buffer is played and the workflow updates to current state DAWG

    def __init__(self, short_id, parent_path, input_key=None, long_id=None):
        self.id = short_id
        self.key = input_key
        self.long_id = long_id
        self.replay_buffer = []
        self.started = long_id is not None
        self.calls = []
        self.last_message = ''
        self.path = None
        self.parent_path = parent_path
        self.failure = None

    @property
    def status(self):
        """
        Property. Get the status of the most recent Call to start
        """
        if len(self.calls):
            status = self.calls[-1].status
            if status == '-':
                return 'Starting'
            return status
        return 'Pending'

    def handle(self, evt, *args, **kwargs):
        if not self.started:
            self.replay_buffer.append((evt, args, kwargs))
            return
        attribute = 'on_'+evt
        getattr(self, attribute)(*args, **kwargs)

    def on_start(self, input_key, long_id):
        # print("Handling start event")
        self.started = True
        self.long_id = long_id
        self.input_key = input_key
        for event, args, kwargs in self.replay_buffer:
            # print("Replaying previous events...")
            self.handle(event, *args, **kwargs)
        self.replay_buffer = []

    def on_task(self, workflow, task, na, call, operation):
        # print("Starting task", workflow, task, na, call, operation)
        path = os.path.join(
            self.parent_path,
            'workspace',
            workflow,
            self.long_id,
            'call-'+task
        )
        if call > 1:
            path = os.path.join(path, 'attempt-'+str(call))
        self.calls.append(Call(
            path,
            task,
            call,
            operation
        ))

    def on_message(self, message):
        pass
        # if len(self.calls):
        #     self.calls[-1].last_message = message.strip()
        # else:
            # print("Discard message", message)

    def on_fail(self, message, status=None):
        self.failure = message

    def on_status(self, task, attempt, old, new):
        for call in self.calls:
            if call.task == task and call.attempt == attempt:
                call.status = new
                return
        # else:
            # print("Discard status", old,'->', new)

    # def abort(self):
    #     for call in self.calls:
    #         print("Aborting", call.operation)
    #         abort_operation(call)
    #         status = get_operation_status(call.operation, False)
    #         result = instance_name_pattern.search(status)
    #         if result:
    #             kill_machines(result.group(1).strip())
        # if len(self.calls):
        #     # print("Aborting", self.calls[-1].operation)
        #     abort_operation(self.calls[-1].operation)
        # if self.long_id is not None:
        #     kill_machines('cromwell-workflow-id=cromwell-'+self.long_id)



# wf = WFAdapter(input_key, short_id, long_id=None)
# wf.handle(event)
# ...
# wf.handle(start_event)
# |
# |_ [self.handle(event) for event in self.replay_buffer]
